options(contrasts=c("contr.sum","contr.poly"))
options(warn=-1)

options(scipen=999)
options(width=256)

library("gdata")


library("ROCR")
library("pROC")
library("randomForest")
library("caret")
library("xgboost")
library("e1071")

library("pracma")


# clear all
rm(list=ls())


min_importance = 1

d = read.csv("./database.csv",sep=",")

d$class = factor(d$class)

Niterations = 20 ### HOW MANY VARIABLES
Nsplit = dim(d)[1]	 ### HOW MANY train/test splits for each iteration
seed = 42

set.seed(seed)

n_show = 20

fitControl <- trainControl(
  method = 'LOOCV',                                   
  savePredictions = 'final',        
  classProbs = TRUE,
  summaryFunction=twoClassSummary 
) 

# choose features with the highest importance
writeLines("Evaluating feature importances...")

X = d[,c(3:(dim(d)[2]))]

imp <- readRDS("importance_XGB_ALL.rds")
bvars = rownames(imp$importance)
print(cbind(bvars[c(1:n_show)],round(imp$importance[c(1:n_show),],2)))

best_acc = 0
best_mdl = {}
best_no_of_vars = 0
best_vars = {}
BEST_MODELS = {}

writeLines("Training models...")
j = 2
while(j<=Niterations){
	# tune hyperparameters
	nrounds <- 1000
	# 1. start from number of iterations and learning rate
	writeLines("Tuning round 1: n.iterations and learning rate")
	tune_grid1 <- expand.grid(
	  nrounds = seq(from = 200, to = nrounds, by = 50),
	  eta = c(0.025, 0.05, 0.1, 0.3),
	  max_depth = c(2, 3, 4, 5, 6),
	  gamma = 0,
	  colsample_bytree = 1,
	  min_child_weight = 1,
	  subsample = 1
	)
	xgb_model <- train(class~., data=d[c("class",bvars[1:j])], method='xgbTree', trControl = fitControl, tuneGrid = tune_grid1, nthread=4, verbose =FALSE, allowParallel=TRUE)
	best_eta =  xgb_model$bestTune$eta
	max_depth = xgb_model$bestTune$max_depth

	# 2. Maximum Depth and Minimum Child Weight
	writeLines("Tuning round 2: max.depth and min.child weight")
	tune_grid2 <- expand.grid(
	  nrounds = seq(from = 50, to = nrounds, by = 50),
	  eta = best_eta,
	  max_depth = ifelse(max_depth == 2,
		c(max_depth:4),
		max_depth-1:max_depth+1),
	  gamma = 0,
	  colsample_bytree = 1,
	  min_child_weight = c(1, 2, 3),
	  subsample = 1
	)
	xgb_model <- train(class~., data=d[c("class",bvars[1:j])], method='xgbTree', trControl = fitControl, tuneGrid = tune_grid2, nthread=4, verbose =FALSE, allowParallel=TRUE)
	best_max_depth = xgb_model$bestTune$max_depth
	best_min_child_weight = xgb_model$bestTune$min_child_weight

	# 3. Column and Row Sampling
	writeLines("Tuning round 3: col. and row. sampling")
	tune_grid3 <- expand.grid(
	  nrounds = seq(from = 50, to = nrounds, by = 50),
	  eta = best_eta,
	  max_depth = best_max_depth,
	  gamma = 0,
	  colsample_bytree = c(0.4, 0.6, 0.8, 1.0),
	  min_child_weight = best_min_child_weight,
	  subsample = c(0.5, 0.75, 1.0)
	)
	xgb_model <- train(class~., data=d[c("class",bvars[1:j])], method='xgbTree', trControl = fitControl, tuneGrid = tune_grid3, nthread=4, verbose =FALSE, allowParallel=TRUE)
	best_colsample = xgb_model$bestTune$colsample_bytree
	best_subsample = xgb_model$bestTune$subsample

	# 4. Gamma
	writeLines("Tuning round 4: gamma")
	tune_grid4 <- expand.grid(
	  nrounds = seq(from = 50, to = nrounds, by = 50),
	  eta = best_eta,
	  max_depth = best_max_depth,
	  gamma = c(0, 0.05, 0.1, 0.5, 0.7, 0.9, 1.0),
	  colsample_bytree = best_colsample,
	  min_child_weight = best_min_child_weight,
	  subsample = best_subsample
	)
	xgb_model <- train(class~., data=d[c("class",bvars[1:j])], method='xgbTree', trControl = fitControl, tuneGrid = tune_grid4, nthread=4, verbose =FALSE, allowParallel=TRUE)
	best_gamma = xgb_model$bestTune$gamma

	# 5. Reducing the Learning Rate
	writeLines("Tuning round 5: reducing learning rate")
	tune_grid5 <- expand.grid(
	  nrounds = seq(from = 100, to = 10000, by = 100),
	  eta = c(0.01, 0.015, 0.025, 0.05, 0.1),
	  max_depth = best_max_depth,
	  gamma = best_gamma,
	  colsample_bytree = best_colsample,
	  min_child_weight = best_min_child_weight,
	  subsample = best_subsample
	)
	xgb_model <- train(class~., data=d[c("class",bvars[1:j])], method='xgbTree', trControl = fitControl, tuneGrid = tune_grid5, nthread=4, verbose =FALSE, allowParallel=TRUE)
	best_nrounds = xgb_model$bestTune$nrounds
	best_eta = xgb_model$bestTune$eta
	best_max_depth = xgb_model$bestTune$max_depth
	best_gamma = xgb_model$bestTune$gamma
	best_colsample = xgb_model$bestTune$colsample_bytree
	best_min_child_weight = xgb_model$bestTune$min_child_weight
	best_subsample = xgb_model$bestTune$subsample

	# 6. Fitting the model
	writeLines("LAST: Fitting the model")
	final_grid <- expand.grid(
	  nrounds = best_nrounds,
	  eta = best_eta,
	  max_depth = best_max_depth,
	  gamma = best_gamma,
	  colsample_bytree = best_colsample,
	  min_child_weight = best_min_child_weight,
	  subsample = best_subsample
	)
	xgb_model <- train(class~., data=d[c("class",bvars[1:j])], method='xgbTree', trControl = fitControl, tuneGrid = final_grid, nthread=4, verbose =FALSE, allowParallel=TRUE)
	
	cc = confusionMatrix(xgb_model$pred["pred"][,1],xgb_model$pred["obs"][,1],positive="RBD")
	sens = cc$byClass[1]
	spec = cc$byClass[2]
	acc  = cc$overall[1]
	kappa = cc$overall[2]
	acc_low = cc$overall[3]
	acc_up = cc$overall[4]
	fprintf("No. of variables: %d, accuracy=%.2f(%.2f,%.2f), sens=%.2f, spec=%.2f\n",j,acc,acc_low,acc_up,sens,spec)
	if(acc>best_acc){
		best_acc = acc
		best_no_of_vars = j
		best_vars = bvars[c(1:j)]
		#pred = predict(best_model, newdata=d[,c("class",best_vars)], type="response")
		#cc = confusionMatrix(pred,d$class,positive='RBD')
		#fprintf("*** Accuracy on whole training dataset = %.2f\n",cc$overall[1])
		fname = gsub(" ","",paste('model_XGB_mean_',j,'vars.RData'))
		save(xgb_model, best_vars, acc, sens, spec, kappa, file=fname)		
	}	
	j = j+1
}

